"""MCP protocol server implementation."""

from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import glob
from typing import TYPE_CHECKING, Any, Self

import mcp
from mcp.server import NotificationOptions, Server
from mcp.types import (
    INTERNAL_ERROR,
    INVALID_PARAMS,
    Completion,
    CompletionArgument,
    GetPromptResult,
    Resource,
    TextContent,
)

from llmling.config.manager import ConfigManager
from llmling.config.models import PathResource, SourceResource
from llmling.config.runtime import RuntimeConfig
from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.server import conversions
from llmling.server.log import configure_server_logging, run_logging_processor
from llmling.server.observers import PromptObserver, ResourceObserver, ToolObserver


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Coroutine
    import os

    from pydantic import AnyUrl

    from llmling.prompts.models import Prompt


logger = get_logger(__name__)


class LLMLingServer:
    """MCP protocol server implementation."""

    def __init__(
        self,
        runtime: RuntimeConfig,
        *,
        name: str = "llmling-server",
    ) -> None:
        """Initialize server with runtime configuration.

        Args:
            runtime: Fully initialized runtime configuration
            name: Server name for MCP protocol
        """
        self.name = name
        self.runtime = runtime

        # Create MCP server
        self.server = Server(name)
        self.server.notification_options = NotificationOptions(
            prompts_changed=True,
            resources_changed=True,
            tools_changed=True,
        )
        self._tasks: set[asyncio.Task[Any]] = set()

        self._setup_handlers()
        self._setup_observers()

    @classmethod
    @asynccontextmanager
    async def from_config_file(
        cls,
        config_path: str | os.PathLike[str],
        *,
        name: str = "llmling-server",
    ) -> AsyncGenerator[LLMLingServer, None]:
        """Create and run server from config file with proper context management.

        Args:
            config_path: Path to configuration file
            name: Optional server name

        Example:
            ```python
            async with LLMLingServer.from_config_file("config.yml") as server:
                await server.start()
            ```
        """
        manager = ConfigManager.load(config_path)
        async with RuntimeConfig.from_config(manager.config) as runtime:
            server = cls(runtime, name=name)
            try:
                yield server
            finally:
                await server.shutdown()

    def _create_task(self, coro: Coroutine[None, None, Any]) -> asyncio.Task[Any]:
        """Create and track an asyncio task."""
        task: asyncio.Task[Any] = asyncio.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    def _setup_handlers(self) -> None:
        """Register MCP protocol handlers."""

        @self.server.set_logging_level()
        async def handle_set_level(level: mcp.LoggingLevel) -> None:
            """Handle logging level changes."""
            try:
                python_level = conversions.LOG_LEVEL_MAP[level]
                logger.setLevel(python_level)
                data = f"Log level set to {level}"
                await self.current_session.send_log_message(level, data, logger=self.name)
            except Exception as exc:
                error = mcp.McpError("Error setting log level")
                error.error = mcp.ErrorData(code=INTERNAL_ERROR, message=str(exc))
                raise error from exc

        @self.server.list_tools()
        async def handle_list_tools() -> list[mcp.types.Tool]:
            """Handle tools/list request."""
            return [conversions.to_mcp_tool(tool) for tool in self.runtime.get_tools()]

        @self.server.call_tool()
        async def handle_call_tool(
            name: str,
            arguments: dict[str, Any] | None = None,
        ) -> list[TextContent]:
            """Handle tools/call request."""
            arguments = arguments or {}
            # Filter out _meta from arguments
            args = {k: v for k, v in arguments.items() if not k.startswith("_")}
            try:
                result = await self.runtime.execute_tool(name, **args)
                return [TextContent(type="text", text=str(result))]
            except Exception as exc:
                logger.exception("Tool execution failed: %s", name)
                error_msg = f"Tool execution failed: {exc}"
                return [TextContent(type="text", text=error_msg)]

        @self.server.list_prompts()
        async def handle_list_prompts() -> list[mcp.types.Prompt]:
            """Handle prompts/list request."""
            return [conversions.to_mcp_prompt(p) for p in self.runtime.get_prompts()]

        @self.server.get_prompt()
        async def handle_get_prompt(
            name: str,
            arguments: dict[str, str] | None = None,
        ) -> GetPromptResult:
            """Handle prompts/get request."""
            try:
                prompt = self.runtime.get_prompt(name)
                messages = await prompt.format(arguments or {})  # Note: now async
                mcp_msgs = [conversions.to_mcp_message(msg) for msg in messages]
                return GetPromptResult(description=prompt.description, messages=mcp_msgs)
            except exceptions.LLMLingError as exc:
                msg = str(exc)
                error = mcp.McpError(msg)
                code = INVALID_PARAMS if "not found" in msg else INTERNAL_ERROR
                error.error = mcp.ErrorData(code=code, message=msg)
                raise error from exc

        @self.server.list_resources()
        async def handle_list_resources() -> list[Resource]:
            """Handle resources/list request."""
            resources = []
            for name in self.runtime.list_resources():
                try:
                    # First get URI and basic info without loading
                    uri = self.runtime.get_resource_uri(name)
                    mcp_uri = conversions.to_mcp_uri(uri)
                    dsc = self.runtime._config.resources[name].description
                    mime = "text/plain"  # Default, could be made more specific
                    res = Resource(uri=mcp_uri, name=name, description=dsc, mimeType=mime)
                    resources.append(res)
                except Exception:
                    msg = "Failed to create resource listing for %r. Config: %r"
                    logger.exception(msg, name, self.runtime._config.resources.get(name))
                    continue

            return resources

        @self.server.read_resource()
        async def handle_read_resource(uri: mcp.types.AnyUrl) -> str | bytes:
            """Handle direct resource content requests."""
            try:
                internal_uri = conversions.from_mcp_uri(str(uri))
                logger.debug("Loading resource from internal URI: %s", internal_uri)

                if "://" not in internal_uri:
                    resource = await self.runtime.load_resource(internal_uri)
                else:
                    resource = await self.runtime.load_resource_by_uri(internal_uri)

                if resource.metadata.mime_type.startswith("text/"):
                    return resource.content
                return resource.content_items[0].content.encode()

            except Exception as exc:
                msg = f"Failed to read resource: {exc}"
                logger.exception(msg)
                error = mcp.McpError(msg)
                error.error = mcp.ErrorData(code=INTERNAL_ERROR, message=str(exc))
                raise error from exc

        @self.server.completion()
        async def handle_completion(
            ref: mcp.types.PromptReference | mcp.types.ResourceReference,
            argument: mcp.types.CompletionArgument,
        ) -> mcp.types.Completion:
            """Handle completion requests."""
            try:
                match ref:
                    case mcp.types.PromptReference():
                        values = await self.runtime.get_prompt_completions(
                            current_value=argument.value,
                            argument_name=argument.name,
                            prompt_name=ref.name,
                        )
                    case mcp.types.ResourceReference():
                        values = await self.runtime.get_resource_completions(
                            uri=ref.uri,
                            current_value=argument.value,
                            argument_name=argument.name,
                        )
                    case _:
                        msg = f"Invalid reference type: {type(ref)}"
                        raise ValueError(msg)  # noqa: TRY301

                return mcp.types.Completion(
                    values=values[:100],
                    total=len(values),
                    hasMore=len(values) > 100,  # noqa: PLR2004
                )
            except Exception:
                logger.exception("Completion failed")
                return mcp.types.Completion(values=[], total=0, hasMore=False)

        @self.server.progress_notification()
        async def handle_progress(
            token: str | int,
            progress: float,
            total: float | None,
        ) -> None:
            """Handle progress notifications from client."""
            msg = "Progress notification: %s %.1f/%.1f"
            logger.debug(msg, token, progress, total or 0.0)

    def _setup_observers(self) -> None:
        """Set up registry observers for MCP notifications."""
        self.resource_observer = ResourceObserver(self)
        self.prompt_observer = PromptObserver(self)
        self.tool_observer = ToolObserver(self)

        self.runtime.add_resource_observer(self.resource_observer.events)
        self.runtime.add_prompt_observer(self.prompt_observer.events)
        self.runtime.add_tool_observer(self.tool_observer.events)

    async def start(self, *, raise_exceptions: bool = False) -> None:
        """Start the server."""
        try:
            # Start MCP server
            handler = configure_server_logging(self.server)
            options = self.server.create_initialization_options()
            async with (
                mcp.stdio_server() as (read_stream, write_stream),
                asyncio.TaskGroup() as tg,
            ):
                tg.create_task(run_logging_processor(handler))
                await self.server.run(
                    read_stream,
                    write_stream,
                    options,
                    raise_exceptions=raise_exceptions,
                )
        finally:
            await self.shutdown()

    async def shutdown(self) -> None:
        """Shutdown the server."""
        try:
            # Cancel all pending tasks
            if self._tasks:
                for task in self._tasks:
                    task.cancel()
                await asyncio.gather(*self._tasks, return_exceptions=True)
                self._tasks.clear()

            # Remove observers
            self.runtime.remove_resource_observer(self.resource_observer.events)
            self.runtime.remove_prompt_observer(self.prompt_observer.events)
            self.runtime.remove_tool_observer(self.tool_observer.events)

            # Shutdown runtime
            await self.runtime.shutdown()
        finally:
            self._tasks.clear()

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        return self

    async def __aexit__(self, *exc: object) -> None:
        """Async context manager exit."""
        await self.shutdown()

    @property
    def current_session(self) -> mcp.ServerSession:
        """Get current session from request context."""
        try:
            return self.server.request_context.session
        except LookupError as exc:
            msg = "No active request context"
            raise RuntimeError(msg) from exc

    def notify_progress(
        self,
        token: str,
        progress: float,
        total: float | None = None,
        description: str | None = None,
    ) -> None:
        """Send progress notification to client."""
        try:
            # Get current session
            session = self.current_session

            # Create and track the progress notification task
            task = session.send_progress_notification(
                progress_token=token,
                progress=progress,
                total=total,
            )
            self._create_task(task)

            # Optionally send description as log message
            if description:
                coro = session.send_log_message(level="info", data=description)
                self._create_task(coro)

        except Exception:
            logger.exception("Failed to send progress notification")

    async def notify_resource_list_changed(self) -> None:
        """Notify clients about resource list changes."""
        try:
            await self.current_session.send_resource_list_changed()
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send resource list change notification")

    async def notify_resource_change(self, uri: str) -> None:
        """Notify clients about resource changes."""
        try:
            url = conversions.to_mcp_uri(uri)
            self._create_task(self.current_session.send_resource_updated(url))
            self._create_task(self.current_session.send_resource_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send resource change notification")

    async def notify_prompt_list_changed(self) -> None:
        """Notify clients about prompt list changes."""
        try:
            self._create_task(self.current_session.send_prompt_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send prompt list change notification")

    async def notify_tool_list_changed(self) -> None:
        """Notify clients about tool list changes."""
        try:
            self._create_task(self.current_session.send_tool_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send tool list change notification")

    async def _complete_prompt_argument(
        self,
        prompt: Prompt,
        arg_name: str,
        current_value: str,
    ) -> Completion:
        """Generate completions for prompt arguments."""
        try:
            # Get completions through runtime
            completions = await self.runtime.get_prompt_completions(
                prompt.name, arg_name, current_value
            )

            # Add any available defaults if no current value
            arg = next((a for a in prompt.arguments if a.name == arg_name), None)
            if arg and not current_value:
                if arg.default is not None:
                    completions.append(str(arg.default))

                # Add description-based suggestions
                if arg.description and "one of:" in arg.description:
                    try:
                        options_part = arg.description.split("one of:", 1)[1]
                        options = [opt.strip() for opt in options_part.split(",")]
                        completions.extend(
                            opt for opt in options if opt.startswith(current_value)
                        )
                    except IndexError:
                        pass

            # Deduplicate while preserving order
            seen = set()
            unique_completions = [
                x
                for x in completions
                if not (x in seen or seen.add(x))  # type: ignore
            ]

            return Completion(
                values=unique_completions[:100],
                total=len(unique_completions),
                hasMore=len(unique_completions) > 100,  # noqa: PLR2004
            )

        except Exception:
            logger.exception(
                "Completion failed for prompt=%s argument=%s", prompt.name, arg_name
            )
            return Completion(values=[], total=0, hasMore=False)

    async def _complete_resource(
        self,
        uri: AnyUrl,
        argument: CompletionArgument,
    ) -> Completion:
        """Generate completions for resource fields."""
        try:
            # Convert AnyUrl to string for resource lookup
            str_uri = str(uri)
            resource = await self.runtime.load_resource_by_uri(str_uri)
            values: list[str] = []

            # Different completion logic based on resource type
            match resource:
                case PathResource():
                    # Complete paths
                    pattern = f"{argument.value}*"
                    values = list(glob.glob(pattern))  # noqa: PTH207

                case SourceResource():
                    # Complete Python import paths
                    names = self._get_importable_names()
                    values = [name for name in names if name.startswith(argument.value)]

            return Completion(
                values=values[:100],
                total=len(values),
                hasMore=len(values) > 100,  # noqa: PLR2004
            )

        except Exception:
            logger.exception("Resource completion failed")
            return Completion(values=[], total=0, hasMore=False)


if __name__ == "__main__":
    import asyncio
    import sys

    from llmling import config_resources

    config_path = sys.argv[1] if len(sys.argv) > 1 else config_resources.TEST_CONFIG
    # asyncio.run(serve(config_path))



"""Server entry point."""

from __future__ import annotations

import asyncio
import sys

from llmling.core.log import get_logger
from llmling.server.factory import create_runtime_config, create_server


logger = get_logger(__name__)

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


async def main() -> None:
    """Run the LLMling server."""
    config_path = sys.argv[1] if len(sys.argv) > 1 else None

    try:
        runtime = create_runtime_config(config_path)
        async with runtime as r:
            server = create_server(r)
            await server.start(raise_exceptions=True)
    except KeyboardInterrupt:
        logger.info("Server stopped by user")
    except Exception:
        logger.exception("Fatal server error")
        sys.exit(1)


def run() -> None:
    """Entry point for the server."""
    asyncio.run(main())


if __name__ == "__main__":
    run()



from llmling.prompts.models import (
    ExtendedPromptArgument,
    Prompt,
    PromptMessage,
)
from llmling.prompts.registry import PromptRegistry
from llmling.prompts.function import create_prompt_from_callable

__all__ = [
    "ExtendedPromptArgument",
    "Prompt",
    "PromptMessage",
    "PromptRegistry",
    "create_prompt_from_callable",
]



from __future__ import annotations

import pytest

from llmling.core import exceptions
from llmling.prompts.models import ExtendedPromptArgument, Prompt, PromptMessage


@pytest.mark.asyncio
async def test_render_prompt(runtime_config):
    """Test prompt rendering through runtime config."""
    prompt = Prompt(
        name="test",
        description="Test prompt",
        messages=[PromptMessage(role="user", content="Hello {name}")],
        arguments=[ExtendedPromptArgument(name="name", required=True)],
    )
    runtime_config._prompt_registry["test"] = prompt

    messages = await runtime_config.render_prompt("test", {"name": "World"})
    assert len(messages) == 1
    assert messages[0].get_text_content() == "Hello World"


async def test_render_prompt_not_found(runtime_config):
    """Test error handling for non-existent prompts."""
    with pytest.raises(exceptions.LLMLingError, match="Item not found"):
        await runtime_config.render_prompt("nonexistent")


@pytest.mark.asyncio
async def test_render_prompt_validation_error(runtime_config):
    """Test error handling for invalid arguments."""
    prompt = Prompt(
        name="test",
        description="Test prompt",
        messages=[PromptMessage(role="user", content="Hello {name}")],
        arguments=[ExtendedPromptArgument(name="name", required=True)],
    )
    runtime_config._prompt_registry["test"] = prompt

    with pytest.raises(exceptions.LLMLingError, match="Missing required argument"):
        await runtime_config.render_prompt("test", {})



"""Extension system for loading MCP components from entry points."""

from llmling.extensions.base import BaseExtensionLoader
from llmling.extensions.loaders import ToolsetLoader


__all__ = [
    "BaseExtensionLoader",
    "ToolsetLoader",
]



"""Runtime configuration handling.

This module provides the RuntimeConfig class which represents the fully initialized,
"live" state of a configuration, managing all runtime components and registries.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

import depkit
import logfire

from llmling.config.models import PathResource, Prompt, PromptConfig
from llmling.core import exceptions
from llmling.core.events import EventEmitter
from llmling.core.log import get_logger
from llmling.extensions.loaders import ToolsetLoader
from llmling.processors.registry import ProcessorRegistry
from llmling.prompts.function import create_prompt_from_callable
from llmling.prompts.registry import PromptRegistry
from llmling.resources import ResourceLoaderRegistry
from llmling.resources.loaders.path import PathResourceLoader
from llmling.resources.registry import ResourceRegistry
from llmling.tools.base import LLMCallableTool
from llmling.tools.registry import ToolRegistry


if TYPE_CHECKING:
    from collections.abc import Sequence
    import types

    from llmling.config.models import Config, Resource
    from llmling.core.events import RegistryEvents
    from llmling.prompts.models import PromptMessage
    from llmling.resources.models import LoadedResource


logger = get_logger(__name__)


class RuntimeConfig(EventEmitter):
    """Fully initialized runtime configuration.

    This represents the "live" state of a Config, with all components
    initialized and ready to use. It provides a clean interface to
    access and manage runtime resources without exposing internal registries.
    """

    def __init__(
        self,
        config: Config,
        *,
        loader_registry: ResourceLoaderRegistry,
        processor_registry: ProcessorRegistry,
        resource_registry: ResourceRegistry,
        prompt_registry: PromptRegistry,
        tool_registry: ToolRegistry,
    ) -> None:
        """Initialize with config and registries.

        Args:
            config: Original static configuration
            loader_registry: Registry for resource loaders
            processor_registry: Registry for content processors
            resource_registry: Registry for resources
            prompt_registry: Registry for prompts
            tool_registry: Registry for tools
        """
        super().__init__()
        self._config = config
        self._loader_registry = loader_registry
        self._processor_registry = processor_registry
        self._resource_registry = resource_registry
        self._prompt_registry = prompt_registry
        self._tool_registry = tool_registry
        self._initialized = False

        settings = self._config.global_settings
        self._dep_manager = depkit.DependencyManager(
            prefer_uv=settings.prefer_uv,
            requirements=settings.requirements,
            extra_paths=settings.extra_paths,
            pip_index_url=settings.pip_index_url,
            scripts=settings.scripts,
        )

    def __enter__(self) -> Self:
        """Synchronous context manager entry."""
        self._dep_manager.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Synchronous context manager exit."""
        self._dep_manager.__exit__(exc_type, exc_val, exc_tb)

    async def __aenter__(self) -> Self:
        """Initialize dependencies and registries."""
        await self._dep_manager.__aenter__()
        if not self._initialized:
            await self._initialize_registries()
            self._initialized = True
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Clean up dependencies and registries."""
        try:
            await self.shutdown()
        finally:
            await self._dep_manager.__aexit__(exc_type, exc_val, exc_tb)

    def _register_default_components(self) -> None:
        """Register all default components and config items."""
        from llmling.resources import (
            CallableResourceLoader,
            CLIResourceLoader,
            ImageResourceLoader,
            PathResourceLoader,
            SourceResourceLoader,
            TextResourceLoader,
        )

        self._loader_registry["path"] = PathResourceLoader
        self._loader_registry["text"] = TextResourceLoader
        self._loader_registry["cli"] = CLIResourceLoader
        self._loader_registry["source"] = SourceResourceLoader
        self._loader_registry["callable"] = CallableResourceLoader
        self._loader_registry["image"] = ImageResourceLoader

        for name, proc_config in self._config.context_processors.items():
            self._processor_registry[name] = proc_config

        for name, resource in self._config.resources.items():
            self._resource_registry[name] = resource

        for name, tool_config in self._config.tools.items():
            tool = LLMCallableTool.from_callable(
                tool_config.import_path,
                name_override=tool_config.name,
                description_override=tool_config.description,
            )
            self._tool_registry[name] = tool

        if self._config.toolsets:
            loader = ToolsetLoader()
            for name, tool in loader.load_items(self._config.toolsets).items():
                if name not in self._tool_registry:
                    self._tool_registry[name] = tool
                else:
                    logger.warning(
                        "Tool %s from toolset overlaps with configured tool",
                        name,
                    )

        for name, prompt_config in self._config.prompts.items():
            match prompt_config:
                case Prompt():
                    self._prompt_registry[name] = prompt_config
                case PromptConfig():
                    prompt = create_prompt_from_callable(
                        prompt_config.import_path,
                        name_override=prompt_config.name or name,
                        description_override=prompt_config.description,
                        template_override=prompt_config.template,
                    )
                    self._prompt_registry[name] = prompt

    async def _initialize_registries(self) -> None:
        """Initialize all registries."""
        self._register_default_components()
        await self.startup()

    @classmethod
    async def create(cls, config: Config) -> Self:
        """Create and initialize a runtime configuration.

        This is a convenience method that ensures proper initialization
        when not using the async context manager.

        Args:
            config: Static configuration to initialize from

        Returns:
            Initialized runtime configuration
        """
        runtime = cls.from_config(config)
        async with runtime as initialized:
            return initialized

    @classmethod
    @logfire.instrument("Creating runtime configuration")
    def from_config(cls, config: Config) -> Self:
        """Create a fully initialized runtime config from static config.

        Args:
            config: Static configuration to initialize from

        Returns:
            Initialized runtime configuration
        """
        loader_registry = ResourceLoaderRegistry()
        processor_registry = ProcessorRegistry()
        resource_registry = ResourceRegistry(
            loader_registry=loader_registry,
            processor_registry=processor_registry,
        )
        prompt_registry = PromptRegistry()
        tool_registry = ToolRegistry()

        return cls(
            config=config,
            loader_registry=loader_registry,
            processor_registry=processor_registry,
            resource_registry=resource_registry,
            prompt_registry=prompt_registry,
            tool_registry=tool_registry,
        )

    async def startup(self) -> None:
        """Start all runtime components."""
        await self._processor_registry.startup()
        await self._tool_registry.startup()
        await self._resource_registry.startup()
        await self._prompt_registry.startup()

    async def shutdown(self) -> None:
        """Shut down all runtime components."""
        await self._prompt_registry.shutdown()
        await self._resource_registry.shutdown()
        await self._tool_registry.shutdown()
        await self._processor_registry.shutdown()

    # Resource Management
    async def load_resource(self, name: str) -> LoadedResource:
        """Load a resource by name.

        Args:
            name: Name of the resource to load

        Returns:
            Loaded resource content and metadata

        Raises:
            ResourceError: If resource cannot be loaded
        """
        return await self._resource_registry.load(name)

    async def resolve_resource_uri(self, uri_or_name: str) -> tuple[str, Resource]:
        """Resolve a resource identifier to a proper URI and resource.

        Args:
            uri_or_name: Can be:
                - Resource name: "test.txt"
                - Full URI: "file:///test.txt"
                - Local path: "/path/to/file.txt"

        Returns:
            Tuple of (resolved URI, resource object)

        Raises:
            ResourceError: If resolution fails
        """
        logger.debug("Resolving resource identifier: %s", uri_or_name)

        # 1. If it's already a URI, use directly
        if "://" in uri_or_name:
            logger.debug("Using direct URI")
            loader = self._loader_registry.find_loader_for_uri(uri_or_name)
            name = loader.get_name_from_uri(uri_or_name)
            if name in self._resource_registry:
                return uri_or_name, self._resource_registry[name]
            # Create temporary resource for the URI
            resource: Resource = PathResource(path=uri_or_name)
            return uri_or_name, resource

        # 2. Try as resource name
        try:
            logger.debug("Trying as resource name")
            resource = self._resource_registry[uri_or_name]
            loader = self._loader_registry.get_loader(resource)
            uri = loader.create_uri(name=uri_or_name)
        except KeyError:
            pass
        else:
            return uri, resource

        # 3. If it looks like a path, try as file
        if "/" in uri_or_name or "\\" in uri_or_name or "." in uri_or_name:
            try:
                logger.debug("Trying as file path")
                uri = PathResourceLoader.create_uri(name=uri_or_name)
                resource = PathResource(path=uri_or_name)
            except Exception as exc:  # noqa: BLE001
                logger.debug("Failed to create file URI: %s", exc)
            else:
                return uri, resource

        msg = (
            f"Could not resolve resource {uri_or_name!r}. Expected resource name or path."
        )
        raise exceptions.ResourceError(msg)

    async def load_resource_by_uri(self, uri: str) -> LoadedResource:
        """Load a resource by URI."""
        try:
            resolved_uri, resource = await self.resolve_resource_uri(uri)
            loader = self._loader_registry.get_loader(resource)
            loader = loader.create(resource, loader.get_name_from_uri(resolved_uri))
            return await loader.load(processor_registry=self._processor_registry)
        except Exception as exc:
            msg = f"Failed to load resource from URI {uri}"
            raise exceptions.ResourceError(msg) from exc

    def list_resources(self) -> Sequence[str]:
        """List all available resource names.

        Returns:
            List of registered resource names
        """
        return self._resource_registry.list_items()

    def get_resource_uri(self, name: str) -> str:
        """Get URI for a resource.

        Args:
            name: Name of the resource

        Returns:
            URI for the resource

        Raises:
            ResourceError: If resource not found
        """
        return self._resource_registry.get_uri(name)

    def register_resource(
        self,
        name: str,
        resource: Resource,
        *,
        replace: bool = False,
    ) -> None:
        """Register a new resource.

        Args:
            name: Name for the resource
            resource: Resource to register
            replace: Whether to replace existing resource

        Raises:
            ResourceError: If name exists and replace=False
        """
        self._resource_registry.register(name, resource, replace=replace)

    def get_resource_loader(self, resource: Resource) -> Any:  # type: ignore[return]
        """Get loader for a resource type.

        Args:
            resource: Resource to get loader for

        Returns:
            Resource loader instance

        Raises:
            LoaderError: If no loader found for resource type
        """
        return self._loader_registry.get_loader(resource)

    # Tool Management
    def list_tools(self) -> Sequence[str]:
        """List all available tool names.

        Returns:
            List of registered tool names
        """
        return self._tool_registry.list_items()

    @property
    def tools(self) -> dict[str, LLMCallableTool]:
        """Get all registered tools.

        Returns:
            Dictionary mapping tool names to tools
        """
        return dict(self._tool_registry)

    async def execute_tool(self, name: str, **params: Any) -> Any:
        """Execute a tool by name.

        Args:
            name: Name of the tool to execute
            **params: Parameters to pass to the tool

        Returns:
            Tool execution result

        Raises:
            ToolError: If tool execution fails
        """
        return await self._tool_registry.execute(name, **params)

    def get_tool(self, name: str) -> LLMCallableTool:
        """Get a tool by name.

        Args:
            name: Name of the tool

        Returns:
            The tool

        Raises:
            ToolError: If tool not found
        """
        return self._tool_registry[name]

    def get_tools(self) -> Sequence[LLMCallableTool]:
        """Get all registered tools.

        Returns:
            List of all tools
        """
        return list(self._tool_registry.values())

    # Prompt Management
    def list_prompts(self) -> Sequence[str]:
        """List all available prompt names.

        Returns:
            List of registered prompt names
        """
        return self._prompt_registry.list_items()

    async def render_prompt(
        self,
        name: str,
        arguments: dict[str, Any] | None = None,
    ) -> Sequence[PromptMessage]:
        """Format a prompt with arguments.

        Args:
            name: Name of the prompt
            arguments: Optional arguments for formatting

        Returns:
            List of formatted messages

        Raises:
            LLMLingError: If prompt not found or formatting fails
        """
        try:
            prompt = self._prompt_registry[name]
            return await prompt.format(arguments)
        except KeyError as exc:
            msg = f"Prompt not found: {name}"
            raise exceptions.LLMLingError(msg) from exc
        except Exception as exc:
            msg = f"Failed to format prompt {name}: {exc}"
            raise exceptions.LLMLingError(msg) from exc

    def get_prompt(self, name: str) -> Prompt:
        """Get a prompt by name.

        Args:
            name: Name of the prompt

        Returns:
            The prompt

        Raises:
            LLMLingError: If prompt not found
        """
        try:
            return self._prompt_registry[name]
        except KeyError as exc:
            msg = f"Prompt not found: {name}"
            raise exceptions.LLMLingError(msg) from exc

    def get_prompts(self) -> Sequence[Prompt]:
        """Get all registered prompts.

        Returns:
            List of all prompts
        """
        return list(self._prompt_registry.values())

    # Registry Observation
    def add_resource_observer(self, observer: RegistryEvents[str, Resource]) -> None:
        """Add observer for resource changes.

        Args:
            observer: Observer to add
        """
        self._resource_registry.add_observer(observer)

    def add_prompt_observer(self, observer: RegistryEvents[str, Prompt]) -> None:
        """Add observer for prompt changes.

        Args:
            observer: Observer to add
        """
        self._prompt_registry.add_observer(observer)

    def add_tool_observer(self, observer: RegistryEvents[str, LLMCallableTool]) -> None:
        """Add observer for tool changes.

        Args:
            observer: Observer to add
        """
        self._tool_registry.add_observer(observer)

    def remove_resource_observer(self, observer: RegistryEvents[str, Resource]) -> None:
        """Remove resource observer.

        Args:
            observer: Observer to remove
        """
        self._resource_registry.remove_observer(observer)

    def remove_prompt_observer(self, observer: RegistryEvents[str, Prompt]) -> None:
        """Remove prompt observer.

        Args:
            observer: Observer to remove
        """
        self._prompt_registry.remove_observer(observer)

    def remove_tool_observer(
        self, observer: RegistryEvents[str, LLMCallableTool]
    ) -> None:
        """Remove tool observer.

        Args:
            observer: Observer to remove
        """
        self._tool_registry.remove_observer(observer)

    @property
    def original_config(self) -> Config:
        """Get the original static configuration.

        Returns:
            Original configuration
        """
        return self._config

    async def get_prompt_completions(
        self,
        current_value: str,
        argument_name: str,
        prompt_name: str,
        **options: Any,
    ) -> list[str]:
        """Get completions for a prompt argument.

        Args:
            current_value: Current input value
            argument_name: Name of the argument
            prompt_name: Name of the prompt
            **options: Additional options

        Returns:
            List of completion suggestions
        """
        return await self._prompt_registry.get_completions(
            current_value=current_value,
            argument_name=argument_name,
            prompt_name=prompt_name,
            **options,
        )

    async def get_resource_completions(
        self,
        uri: str,
        current_value: str,
        argument_name: str | None = None,
        **options: Any,
    ) -> list[str]:
        """Get completions for a resource.

        Args:
            uri: Resource URI
            current_value: Current input value
            argument_name: Optional argument name
            **options: Additional options

        Returns:
            List of completion suggestions
        """
        loader = self._loader_registry.find_loader_for_uri(uri)
        return await loader.get_completions(
            current_value=current_value,
            argument_name=argument_name,
            **options,
        )


if __name__ == "__main__":
    from llmling import Config

    cfg = Config.from_file("E:/mcp_zed.yml")  # type: ignore
    runtime = RuntimeConfig.from_config(cfg)
    with runtime as ctx:
        pass



"""Convert Python functions to MCP prompts."""

from __future__ import annotations

import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    get_type_hints,
)

from llmling.core.log import get_logger
from llmling.core.typedefs import MessageContent
from llmling.prompts.models import ExtendedPromptArgument, Prompt, PromptMessage


if TYPE_CHECKING:
    from collections.abc import Callable

    from llmling.prompts.completion import CompletionFunction


logger = get_logger(__name__)


def create_prompt_from_callable(
    fn: Callable[..., Any] | str,
    *,
    name_override: str | None = None,
    description_override: str | None = None,
    template_override: str | None = None,
    completions: dict[str, CompletionFunction] | None = None,
) -> Prompt:
    """Create a prompt from a callable.

    Args:
        fn: Function or import path to create prompt from
        name_override: Optional override for prompt name
        description_override: Optional override for prompt description
        template_override: Optional override for message template
        completions: Optional dict mapping argument names to completion functions

    Returns:
        Prompt instance

    Raises:
        ValueError: If callable cannot be imported or is invalid
    """
    # Import if string path provided
    if isinstance(fn, str):
        from llmling.utils import importing

        fn = importing.import_callable(fn)

    # Get function metadata
    name = name_override or fn.__name__
    sig = inspect.signature(fn)
    hints = get_type_hints(fn, include_extras=True)

    # Get description from docstring
    doc = inspect.getdoc(fn)
    description = description_override or (
        doc.split("\n\n")[0] if doc else f"Prompt from {name}"
    )

    # Parse docstring for argument descriptions
    arg_docs = _parse_arg_docs(fn)
    completion_funcs = completions or {}

    # Create arguments
    arguments = []
    for param_name, param in sig.parameters.items():
        if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            continue

        type_hint = hints.get(param_name, Any)
        required = param.default == param.empty
        arg_desc = arg_docs.get(param_name, "")

        arguments.append(
            ExtendedPromptArgument(
                name=param_name,
                description=arg_desc,
                required=required,
                type_hint=type_hint,  # Store original type hint
                default=None if param.default is param.empty else param.default,
                completion_function=completion_funcs.get(param_name),
            )
        )

    # Create message template. Will be formatted with function result
    template = template_override if template_override else "{result}"
    # Create prompt messages
    content = MessageContent(type="text", content=f"Content from {name}:\n")
    messages = [
        PromptMessage(role="system", content=content),
        PromptMessage(role="user", content=MessageContent(type="text", content=template)),
    ]
    path = f"{fn.__module__}.{fn.__qualname__}"
    return Prompt(
        name=name,
        description=description,
        arguments=arguments,
        messages=messages,
        metadata={"source": "function", "import_path": path},
    )


def _parse_arg_docs(fn: Callable[..., Any]) -> dict[str, str]:
    """Parse argument descriptions from docstring.

    Args:
        fn: Function to parse docstring from

    Returns:
        Dictionary mapping argument names to their descriptions
    """
    doc = inspect.getdoc(fn)
    if not doc:
        return {}

    arg_docs: dict[str, str] = {}
    lines = doc.split("\n")
    in_args = False
    current_arg = None

    for line in lines:
        line = line.strip()

        # Start of Args section
        if line == "Args:":
            in_args = True
            continue

        # End of Args section
        if in_args and (not line or line.startswith(("Returns:", "Raises:"))):
            break

        # Parse argument
        if in_args:
            if line and not line.startswith(" "):
                # New argument definition
                if ":" in line:
                    arg_name, desc = line.split(":", 1)
                    current_arg = arg_name.strip()
                    arg_docs[current_arg] = desc.strip()
            elif current_arg and line:
                # Continuation of previous argument description
                arg_docs[current_arg] += " " + line.strip()

    return arg_docs


if __name__ == "__main__":

    def example_function(text: str, mode: Literal["a", "b"] = "a") -> str:
        """Process text with given mode and tags.

        Args:
            text: Input text to process
            mode: Processing mode (one of: a, b)
        """
        return text

    prompt = create_prompt_from_callable(example_function)
    print(f"Created prompt: {prompt}")
    print(f"Arguments: {prompt.arguments}")
    print(f"Messages: {prompt.messages}")



"""Base class for extension loaders."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

from epregistry import EntryPointRegistry

from llmling.core.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Callable


logger = get_logger(__name__)
T = TypeVar("T")


class BaseExtensionLoader[T]:
    """Base class for extension loaders.

    Entry points are expected to be registered under 'llmling' with their type
    as the entry point name, e.g.:
    [project.entry-points.llmling]
    tools = "my_module:get_mcp_tools"
    prompts = "my_module:get_mcp_prompts"
    """

    component_type: ClassVar[str]
    converter: Callable[[Any], T]

    def __init__(self) -> None:
        """Initialize loader."""
        self.registry = EntryPointRegistry[Any]("llmling")

    def load_items(self, module_names: list[str]) -> dict[str, T]:
        """Load items from specified modules."""
        items = {}
        for module in module_names:
            try:
                # Just look for the component type entry point
                if entry_point := self.registry.get(self.component_type):
                    get_items = entry_point.load()
                    for item in get_items():
                        try:
                            converted = self.converter(item)
                            name = getattr(converted, "name", str(item))
                            items[name] = converted
                        except Exception as exc:  # noqa: BLE001
                            logger.warning(
                                "Failed to load item from %s: %s",
                                module,
                                exc,
                            )
            except Exception:
                logger.exception("Failed to load module %s", module)
        return items



from __future__ import annotations

from llmling.extensions.loaders import ToolsetLoader


def test_toolset_loader():
    """Test loading tools from entry points."""
    loader = ToolsetLoader()
    tools = loader.load_items(["llmling"])  # just the package name
    assert "example_tool" in tools
    assert "analyze_ast" in tools



"""Extension loaders for different MCP component types."""

from __future__ import annotations

from llmling.extensions.base import BaseExtensionLoader
from llmling.tools.base import LLMCallableTool


class ToolsetLoader(BaseExtensionLoader[LLMCallableTool]):
    """Loads tools from entry points.

    Entry points should return a list of callable objects:

    def get_mcp_tools() -> list[Callable[..., Any]]:
        return [function1, function2]
    """

    component_type = "tools"
    converter = LLMCallableTool.from_callable



"""Prompt models for MCP."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from llmling.completions import CompletionFunction  # noqa: TC001
from llmling.core.typedefs import MessageContent, MessageRole


class ExtendedPromptArgument(BaseModel):
    """Prompt argument with validation information.

    Extends MCP's prompt argument with additional metadata for validation.
    """

    name: str
    description: str | None = None
    required: bool = False
    type_hint: Any = str
    default: Any | None = None
    completion_function: CompletionFunction = None

    model_config = ConfigDict(frozen=True)


class PromptMessage(BaseModel):
    """A message in a prompt template."""

    role: MessageRole
    content: str | MessageContent | list[MessageContent] = ""

    model_config = ConfigDict(frozen=True)

    def get_text_content(self) -> str:
        """Get text content of message."""
        match self.content:
            case str():
                return self.content
            case MessageContent() if self.content.type == "text":
                return self.content.content
            case list() if self.content:
                # Join text content items with space
                text_items = [
                    item.content
                    for item in self.content
                    if isinstance(item, MessageContent) and item.type == "text"
                ]
                return " ".join(text_items) if text_items else ""
            case _:
                return ""


class Prompt(BaseModel):
    """MCP prompt template."""

    name: str
    description: str
    messages: list[PromptMessage]
    arguments: list[ExtendedPromptArgument] = Field(default_factory=list)
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)

    def validate_arguments(self, provided: dict[str, Any]) -> None:
        """Validate that required arguments are provided."""
        required = {arg.name for arg in self.arguments if arg.required}
        missing = required - set(provided)
        if missing:
            msg = f"Missing required arguments: {', '.join(missing)}"
            raise ValueError(msg)

    async def format(
        self, arguments: dict[str, Any] | None = None
    ) -> list[PromptMessage]:
        """Format prompt messages with arguments.

        Args:
            arguments: Values to format the prompt with

        Returns:
            List of formatted messages

        Raises:
            ValueError: If required arguments are missing
        """
        args = arguments or {}
        self.validate_arguments(args)

        # If this is a function prompt, execute it
        if self.metadata.get("source") == "function":
            try:
                from llmling.utils import calling

                import_path = self.metadata["import_path"]
                result = await calling.execute_callable(import_path, **args)
                format_args = {"result": result}
            except Exception as exc:  # noqa: BLE001
                format_args = {"result": f"Error executing function: {exc}"}
        else:
            # Add default values for optional arguments
            format_args = args.copy()  # Make a copy to avoid modifying input
            for arg in self.arguments:
                if arg.name not in format_args:
                    if arg.default is not None:
                        format_args[arg.name] = arg.default
                    elif not arg.required:
                        format_args[arg.name] = ""  # Empty string for optional args

        # Format all messages
        formatted_messages = []
        for msg in self.messages:
            match msg.content:
                case str():
                    content: MessageContent | list[MessageContent] = MessageContent(
                        type="text", content=msg.content.format(**format_args)
                    )
                case MessageContent() if msg.content.type == "text":
                    content = MessageContent(
                        type="text", content=msg.content.content.format(**format_args)
                    )
                case list():
                    content = [
                        MessageContent(
                            type=item.type,
                            content=item.content.format(**format_args)
                            if item.type == "text"
                            else item.content,
                            alt_text=item.alt_text,
                        )
                        for item in msg.content
                        if isinstance(item, MessageContent)
                    ]
                case _:
                    content = msg.content

            formatted_messages.append(PromptMessage(role=msg.role, content=content))

        return formatted_messages



"""Models for file watching configuration."""

from __future__ import annotations

from pydantic import BaseModel


class WatchConfig(BaseModel):
    """Watch configuration for resources."""

    enabled: bool = False
    """Whether the watch is enabled"""

    patterns: list[str] | None = None
    """List of pathspec patterns (.gitignore style)"""

    ignore_file: str | None = None
    """Path to .gitignore-style file"""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from upath import UPath

from llmling.config.models import PathResource
from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.resources.base import ResourceLoader, create_loaded_resource
from llmling.utils import paths


logger = get_logger(__name__)

if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.resources.models import LoadedResource


class PathResourceLoader(ResourceLoader[PathResource]):
    """Loads context from files or URLs."""

    context_class = PathResource
    uri_scheme = "file"
    supported_mime_types: ClassVar[list[str]] = [
        "text/plain",
        "application/json",
        "text/markdown",
        "text/yaml",
    ]

    @classmethod
    def get_name_from_uri(cls, uri: str) -> str:
        """Extract the normalized path from a URI."""
        try:
            return paths.uri_to_path(uri)
        except ValueError as exc:
            msg = f"Invalid URI: {uri}"
            raise exceptions.LoaderError(msg) from exc

    @classmethod
    def create_uri(cls, *, name: str) -> str:
        """Create a URI from a path."""
        try:
            return paths.path_to_uri(name)
        except ValueError as exc:
            msg = f"Failed to create URI from {name}"
            raise exceptions.LoaderError(msg) from exc

    async def get_completions(
        self,
        current_value: str,
        argument_name: str | None = None,
        **options: Any,
    ) -> list[str]:
        """Get path completions."""
        try:
            # Handle both absolute and relative paths
            path = UPath(current_value) if current_value else UPath()
            if not current_value:
                # Show current directory contents for empty input
                matches = list(path.glob("*"))
            else:
                pattern = f"{path}*" if path.is_dir() else f"{path.parent}/*"
                matches = list(UPath().glob(pattern))

            return [str(p) for p in matches if str(p).startswith(current_value or ".")]
        except Exception:
            logger.exception("Path completion failed")
            return []

    async def _load_impl(
        self,
        resource: PathResource,
        name: str,
        processor_registry: ProcessorRegistry | None,
    ) -> LoadedResource:
        """Load content from a file or URL."""
        try:
            path = UPath(resource.path)
            content = path.read_text("utf-8")

            if processor_registry and (procs := resource.processors):
                processed = await processor_registry.process(content, procs)
                content = processed.content
            meta = {"type": "path", "path": str(path), "scheme": path.protocol}
            return create_loaded_resource(
                content=content,
                source_type="path",
                uri=self.create_uri(name=name),
                mime_type=self.supported_mime_types[0],
                name=resource.description or path.name,
                description=resource.description,
                additional_metadata=meta,
            )
        except Exception as exc:
            msg = f"Failed to load content from {resource.path}"
            raise exceptions.LoaderError(msg) from exc


if __name__ == "__main__":
    uri = PathResourceLoader.create_uri(name="/path/to/file.txt")
    print(uri)  # file:///path/to/file.txt
